import xml.dom.minidom
from xml.dom import *
import unicodedata
import tokenize
import re
import itertools
import math

# Loads a CUED xml file into a python dictionary
class DialogTranscription:
    
    content = {}  # The turns of the dialogue

    def loadNode(self, xmlnode):
        hasText = False
        hasElements = False
        text = ""
        elementAns = {}
        listingValues = []
        
        if xmlnode.nodeType == Node.ELEMENT_NODE:
            for k in xmlnode.attributes.keys():
                elementAns[k] = xmlnode.attributes[k].value
            hasElements = True
        
        for node in xmlnode.childNodes:
            if node.nodeType == Node.TEXT_NODE:
                val = unicodedata.normalize('NFKD', (node.nodeValue)).encode('ascii','ignore') .strip()
                if len(val) > 0:
                    hasText = True
                    text += val
            elif node.nodeType == Node.ELEMENT_NODE:
                val = self.loadNode(node)
                if node.tagName in listingValues:
                    elementAns[node.tagName+"_list"].append(val)
                elif node.tagName in elementAns.keys():
                    elementAns[node.tagName+"_list"] = []
                    elementAns[node.tagName+"_list"].append(elementAns[node.tagName])
                    elementAns[node.tagName+"_list"].append(val)
                    elementAns.pop(node.tagName)
                    listingValues.append(node.tagName)
                else:
                    elementAns[node.tagName] = val
                    
                hasElements = True
        
        if hasElements:
            if hasText:
                elementAns['value'] = text
            return elementAns
        else:
            return text
        
    
    def __init__(self, filename):
        doc = xml.dom.minidom.parse(filename)
        self.content = self.loadNode(doc.documentElement)
        
# Gets a list of items from the dictionary
def getList(key, dict):
    ans = []
    if key + '_list' in dict.keys():
        ans = dict[key + '_list']
    elif key in dict.keys():
        ans.append(dict[key])
    return ans

splitter = re.compile(r'(\w+|\(|\)|,|=|"[^"]*"|\|)')


class ActItem:
    def __init__(self, slot, value):
        self.slot = slot.lower()
        if len(value) > 1 and value[0] == '"' and value[-1] == '"':
            value = value[1:-1]
        self.value = value.lower()
        
    def __str__(self):
        if ' ' in self.value or self.value.isdigit():
            return self.slot + "=\"" + self.value + '"'
        elif len(self.value) == 0:
            return self.slot
        else:
            return self.slot + "=" + self.value
        
    def __hash__(self):
        return hash(self.slot + "=" + self.value)

    def __eq__(self, other):
        return self.slot == other.slot and self.value == other.value
     
# A DialogueAct consists of an act type (representing the speech act type)
# and a list of ActItem's, representing the semantic information
class DialogueAct:
    
    def __init__(self, s="inform()"):
        self.items = []        
        parts = splitter.findall(s)
        if len(parts) > 0:
            self.actType = parts[0]
            
            slot = ""
            val = ""
            
            order = ["slot", "equals", "value", "comma"]
            curOrder = 0
            for i in range(2,len(parts)-1): # Drop brackets
                if parts[i] == "=":
                    curOrder = 2
                elif parts[i] == ",":
                    if len(slot) > 0:
                        self.items.append(ActItem(slot,""))
                    curOrder = 0
                elif order[curOrder] == "slot":
                    slot = parts[i]
                    curOrder == 1
                elif order[curOrder] == "value":
                    val = parts[i]
                    self.items.append(ActItem(slot,val))
                    val = ""
                    slot = ""
                    curOrder = 3
                else:
                    # if we have something here its because the act wasn't well formed
                    # Just add the values onto the last act item
                    if len(self.items) > 0:
                        self.items[-1].value += " " + parts[i]
            if len(slot) > 0:
                self.items.append(ActItem(slot,""))
        else:
            raise BaseException("Invalid act " + str(s))
    
    def __str__(self):
        itemStrings = [str(item) for item in self.items]
        itemStrings.sort()
        return self.actType + "(" + ",".join(itemStrings) + ")"
        
    def subsetOfItems(self, other):
        for item in self.items:
            found = False
            for item2 in other.items:
                if item == item2:
                    found = True
                    break
            if not found:
                return False
        return True
        
    # Returns the number of items in this act that are also in the other
    def numItemsInOther(self, other):
        myItemStrings = [str(item) for item in self.items]
        otherItemStrings = set([str(item) for item in other.items])
        count = 0
        for item in myItemStrings:
            if item in otherItemStrings:
                count += 1
                
        return count
        
    # The distance between two acts
    def distance(self, other):
        a = len(self.items)
        b = len(other.items)
        both = self.numItemsInOther(other)
        extra = 0
        if self.actType != other.actType:
            extra = 1
        if a > b:
            return a - both + extra
        else:
            return b - both + extra
        
    def __eq__(self, other):
        return str(self) == str(other)
        
    def __hash__(self):
        return hash(str(self))

        
# A DialogueUtterance is a list of DialogueActs (represented in a string by
# using the | as a separator). Note that this is not used in the Mar09Trial data
class DialogueUtterance:
    
    def __init__(self, str = ""):
        self.acts = []
        parts = splitter.findall(str)
        if len(parts) > 0:
            cur = ""
            for part in parts:
                if part == "|":
                    self.acts.append(DialogueAct(cur))
                    cur = ""
                else:
                    cur += part
            if len(cur) > 0:
                self.acts.append(DialogueAct(cur))
    
    
    def __str__(self):
        itemStrings = [str(act) for act in self.acts]
        itemStrings.sort()
        return "|".join(itemStrings)
    
    def __hash__(self):
        return hash(str(self))
        
    def __eq__(self, other):
        return str(self) == str(other)
        
    # Returns the max distance between two acts
    def maxdistance(self, other):
        if len(other.acts) == 0:
            if len(self.acts) == 0:
                return 0
            else:
                return 1
        return max([x.distance(y) for (x,y) in itertools.product(self.acts, other.acts)])
        


# Finds the required remaining prob in the list
def getNullProb(hypList):
    total = 0
    for hyp in hypList:
        if 'prob' in hyp:
            hypProb = float(hyp['prob'])
            total += hypProb
        else:
            print "Unable to find prob in " + str(hypList)
    if total < 1:
        return 1 - total
    else:
        return 0

# Gets the map of item --> prob for a list of semihyps    
def getItems(turnoriginal):
    turn = list(turnoriginal)
    ans = {}
    nullProb = getNullProb(turn)
    turn.append({'value' : 'null()', 'prob' : nullProb})
    for semihyp in turn:
        act = DialogueAct(semihyp['value'])
        prob = float(semihyp['prob'])
        if prob > 1E-10:
            items = [str(item) for item in act.items]
            items.append(act.actType)
            for item in items:
                if item in ans:
                    ans[item] += prob
                else:
                    ans[item] = prob
    return ans

# Gets the map of act --> prob for a list of semihyps
def getActs(turnoriginal, addNull = True):
    turn = list(turnoriginal)
    ans = {}
    total = 0
    if addNull:
        nullProb = getNullProb(turn)
        turn.append({'value' : 'null()', 'prob' : nullProb})
    for semihyp in turn:
        act = str(DialogueAct(semihyp['value']))
        prob = float(semihyp['prob'])
        if prob > 1E-10:
            if act in ans.keys():
                ans[act] += prob
            else:
                ans[act] = prob

    sumProb = sum(ans.values())
    if sumProb > 1:
        newVal = {}
        for act, prob in ans.iteritems():
            newVal[act] = prob / sumProb;
        ans = newVal
    return ans

# Gets the map of act --> prob for a list of semihyps
def getAsrHyps(asrhyps):
    asr = list(asrhyps)
    ans = {}
    for hyp in asr:
        if 'value' in hyp:
            a = str(hyp['value'])
            prob = 0
            if 'prob' in hyp:
                prob = float(hyp['prob'])
            elif 'infscore' in hyp:
                prob = math.exp(float(hyp['infscore'])) # Inference evidence is like log prob
            if prob > 1E-10:
                if a in ans.keys():
                    ans[a] += prob
                else:
                    ans[a] = prob

    sumProb = sum(ans.values())
    if sumProb > 1:
        newVal = {}
        for act, prob in ans.iteritems():
            newVal[act] = prob / sumProb;
        ans = newVal
    return ans


# Gets the map of act --> prob for a list of semihyps
def getUtterances(turn, addNull = True):
    turn = list(turn)
    ans = {}
    total = 0
    if addNull:
        nullProb = getNullProb(turn)
        turn.append({'value' : 'null()', 'prob' : nullProb})
    for semihyp in turn:
        act = str(DialogueUtterance(semihyp['value']))
        prob = float(semihyp['prob'])
        if prob > 1E-10:
            if act in ans.keys():
                ans[act] += prob
            else:
                ans[act] = prob
    return ans
