import os, os.path
from cued.extract import *
import random
from sklearn.pipeline import Pipeline
from sklearn.feature_extraction.text import WordNGramAnalyzer, CountVectorizer, ENGLISH_STOP_WORDS
import scipy.sparse
import numpy
from scipy.sparse import csr_matrix
from scipy.io import loadmat, savemat
from scipy.io import mmread, mmwrite

# A class for converting lists of items into a vector of ones and zeros
class ListVectorizer:
    def __init__(self):
        self.map = {}

    def fit_transform(self, lines):
        output = []
        for l in lines:
            newLine = []
            for it in l:
                val = 0
                if it in self.map.keys():
                    val = self.map[it]
                else:
                    val = len(self.map)
                    self.map[it] = val
                newLine.append(val)
            output.append(newLine)
        return output

# A class for converting lists of items into a vector of ones and zeros
class SingleValueAnalyzer:
    def __init__(self):
        self.map = {}

    def fit_transform(self, lines):
        output = []
        for it in lines:
            val = 0
            if it in self.map.keys():
                val = self.map[it]
            else:
                val = len(self.map)
                self.map[it] = val
            output.append(val)
        return output


# Saves the dictionary to a file
def saveDictValues(d, keys, fileName):
    f = open(fileName, "w")
    for k in keys:
        v = d[k]
        f.write(k + " => " + v + "\n")
    f.close()

# Process the corpus and extract the transcriptions, etc
def preprocessCorpus(corpus, asrFile, tranFile, semiFile, uactFile, percentTest):
    xmlfiles = os.listdir(corpus)
    xmlfiles = [os.path.join(corpus, x) for x in xmlfiles if '.xml' in x]    
    asr, tran, semi, uact = extractValuesFromCorpus(xmlfiles)
    
    keys = asr.keys()
    numTest = int(len(keys) * percentTest)
    testKeys = random.sample(keys, numTest)
    trainKeys = [x for x in keys if x not in testKeys]
    
    saveDictValues(asr, testKeys, asrFile + "-test.txt")    
    saveDictValues(tran, testKeys, tranFile + "-test.txt")
    saveDictValues(semi, testKeys, semiFile + "-test.txt")    
    saveDictValues(uact, testKeys, uactFile + "-test.txt")
    
    saveDictValues(asr, trainKeys, asrFile + "-train.txt")    
    saveDictValues(tran, trainKeys, tranFile + "-train.txt")
    saveDictValues(semi, trainKeys, semiFile + "-train.txt")    
    saveDictValues(uact, trainKeys, uactFile + "-train.txt")

# Conver the file of text values into a file with n-gram counts
def convertTextFileToMatrix(fName, vectorizer, mode):
    f = open(fName)
    lines = f.readlines()
    f.close()
    lines = [l.split('=>')[1] for l in lines if len(l) > 0]
    
    print "Vectorising the text"
    if mode == "train":
        X = vectorizer.fit_transform(lines).tocsr()
    else:
        X = vectorizer.transform(lines).tocsr()
    return X

# Convert the file of acts to a matrix with 1s for the items appearing
def convertActFileToMatrix(fName, vectorizer):
    f = open(fName)
    lines = f.readlines()
    f.close()
    lines = [l.split('=>')[1] for l in lines if len(l) > 0]
    lines = [[str(i).lower() for i in DialogueAct(l).items] for l in lines if len(l) > 0]
    vectorized = vectorizer.fit_transform(lines)
    
    Y = csr_matrix((len(lines), len(vectorizer.map)))
    i = 0
    for line in vectorized:
        for val in line:
            Y[i,val] = 1
        i = i + 1
    return Y

# Extract all the types and return them as a vector
def convertActFileToTypeVector(fName, vectorizer):
    f = open(fName)
    lines = f.readlines()
    f.close()
    lines = [l.split('=>')[1] for l in lines if len(l) > 0]
    lines = [DialogueAct(l).actType for l in lines if len(l) > 0]
    Y = vectorizer.fit_transform(lines)
    return numpy.matrix(Y)


if __name__ == '__main__':
    # Run through the dialogues and extract the relevant bits
    preprocessed = "preprocessed"
    corpus = "mar09trial"
    
    if not os.path.isdir(preprocessed):
        os.mkdir(preprocessed)
    
    uactFile = os.path.join(preprocessed, "uact") # Semantic transcriptions
    semiFile = os.path.join(preprocessed, "semi") # Top semantic hypothesis
    tranFile = os.path.join(preprocessed, "transcription") # Orthographic transcriptions
    asrFile  = os.path.join(preprocessed, "asr") # Top ASR hypothesis
    
    #preprocessCorpus(corpus, asrFile, tranFile, semiFile, uactFile, 0.1)
    max_n=3
    ngramiser  = WordNGramAnalyzer(max_n=max_n, stop_words=ENGLISH_STOP_WORDS)
    tranVectorizer = CountVectorizer(analyzer=ngramiser)
    asrVectorizer = CountVectorizer(analyzer=ngramiser)
    actVectorizer = ListVectorizer()
    typeVectorizer = SingleValueAnalyzer()
    
    for mode in ["train", "test"]:
        X = convertTextFileToMatrix(tranFile + "-" + mode +".txt", tranVectorizer, mode)
        savemat(tranFile + '-' + mode + '.mat', {'X' : X})
        mmwrite(tranFile + '-' + mode + '.mtx', X)
        X = convertTextFileToMatrix(asrFile + "-" + mode +".txt", asrVectorizer, mode)
        savemat(asrFile + '-' + mode + '.mat', {'X' : X})
        mmwrite(asrFile + '-' + mode + '.mtx', X)

        Y = convertActFileToMatrix(uactFile + "-" + mode + ".txt", actVectorizer)
        savemat(uactFile + '-' + mode + '.items.mat', {'Y' : Y})
        mmwrite(uactFile + '-' + mode + '.items.mtx', Y)
        Z = convertActFileToTypeVector(uactFile + "-" + mode + ".txt", typeVectorizer)
        savemat(uactFile + '-' + mode + '.types.mat', {'Z' : Z})
        mmwrite(uactFile + '-' + mode + '.types.mtx', Z)
        
        Y = convertActFileToMatrix(semiFile + "-" + mode + ".txt", actVectorizer)
        savemat(semiFile + '-' + mode + '.items.mat', {'Y' : Y})
        mmwrite(semiFile + '-' + mode + '.items.mtx', Y)
        Z = convertActFileToTypeVector(semiFile + "-" + mode + ".txt", typeVectorizer)
        savemat(semiFile + '-' + mode + '.types.mat', {'Z' : Z})
        mmwrite(semiFile + '-' + mode + '.types.mtx', Z)
        print "Done " + mode
